#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/init.h>
#include <linux/moduleparam.h>
#include <linux/rtnetlink.h>
#include <net/rtnetlink.h>
#include <linux/u64_stats_sync.h>

static struct net_device *crypto_dev;
static const char slave_name[]="wlan0";

struct crypto_priv
{
    struct net_device_stats priv_stats;
    struct net_device *slave;
};

rx_handler_result_t handle_frame(struct sk_buff **pskb)
{
    struct sk_buff *skb = *pskb;

    printk(KERN_WARNING "%s: Address src=%pM dst=%pM len=%d data_len=%d mac_len=%d\n", slave_name, eth_hdr(skb)->h_source, eth_hdr(skb)->h_dest, skb->len, skb->data_len, skb->mac_len);

    /* We need to remove scb from slave device and push it to crypto device */

    return RX_HANDLER_PASS;
}

int crypto_dev_init(struct net_device *dev)
{
    return 0;
}

int crypto_open(struct net_device *dev)
{
    printk("%s: device opened\n", dev->name);
    memcpy(dev->dev_addr, "\0ABCD0", ETH_ALEN);
    netif_start_queue(dev);
    return 0;
}

int crypto_close(struct net_device *dev)
{
    printk("%s: device closed\n", dev->name);
    netif_stop_queue(dev);
    return 0;
}

int crypto_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    return 0;
}

netdev_tx_t crypto_xmit(struct sk_buff *skb, struct net_device *dev)
{
    struct crypto_priv *priv = netdev_priv(dev);

    printk("%s: crypto_xmit\n", dev->name);

    priv->priv_stats.tx_packets++;
    priv->priv_stats.tx_bytes += skb->len;

    /* We need to crypt the frame and push in to teh slave device */
#if 0
    if (priv->slave)
    {
        skb->dev = priv->priv_device;
        skb->priority = 1;
        dev_queue_xmit (skb);
        return 0;
    }
#else
    kfree_skb(skb);
#endif

    return NETDEV_TX_OK;
}

struct net_device_stats *crypto_get_stats(struct net_device *dev)
{
    return &((struct crypto_priv *)netdev_priv(dev))->priv_stats;
}

static struct net_device_ops crypto_net_device_ops =
{
    .ndo_init = crypto_dev_init,
    .ndo_open = crypto_open,
    .ndo_stop = crypto_close,
    .ndo_do_ioctl = crypto_ioctl,
    .ndo_get_stats = crypto_get_stats,
    .ndo_start_xmit = crypto_xmit,
};

void crypto_setup(struct net_device *dev)
{
    ether_setup(dev);

    memset(netdev_priv(dev), 0, sizeof(struct crypto_priv));

    dev->netdev_ops = &crypto_net_device_ops;

    random_ether_addr(dev->dev_addr);
}

int __init crypto_init(void)
{
    struct crypto_priv *priv;

    struct net_device *slave;

    int err = 0;

    crypto_dev = alloc_netdev(sizeof(struct crypto_priv), "crypto", crypto_setup);
    if (crypto_dev == NULL) 
        return -ENOMEM;

    slave = __dev_get_by_name(&init_net, slave_name);
    if (!slave)
        return -ENODEV;

    rtnl_lock();

    netdev_rx_handler_register(slave, handle_frame, NULL);

    rtnl_unlock();

    priv = netdev_priv(crypto_dev);

    priv->slave = slave;

    if (( err = dev_alloc_name(crypto_dev, crypto_dev->name) )) {
        printk(KERN_WARNING "%s: allocate name, error %i\n", crypto_dev->name, err);
        return -EIO;
    }

    register_netdev(crypto_dev);

    printk("loaded\n");

    return 0;
}

void __exit crypto_exit(void)
{
    struct net_device *slave;

    rtnl_lock();

    slave = __dev_get_by_name(&init_net, slave_name);

    netdev_rx_handler_unregister(slave);

    rtnl_unlock();

    unregister_netdev(crypto_dev);

    free_netdev(crypto_dev);

    printk("unloaded\n");
}

module_init(crypto_init);
module_exit(crypto_exit);
MODULE_LICENSE("GPL");
